function TimelineController(){
	var self = this; self.name = "TimelineController"; // MANDATORY

    this.data = null;
    this.future_data = false;
    this.conditional = new ConditionalController();
    this.synchronizer = false;
    this.renderer = null;
    this.timer = null;
    this.security_timer = null;

    this.index = 0;
    this.timeline = null;
    this.line = null;
    this.last_time = 0;
    this.toberendered = 0;

    this.template = false;
    this.rendering = false;
    this.scene = false;
    this.pass = 1;
    this.playlist_id = 0;
    this.active = false;

    this.prepared = false;

    this.setDataManager = function(DataManagerController){
        self.data = DataManagerController;
    };

    this.setRenderer = function(DisplayController){
        self.renderer = null;
        self.renderer = DisplayController;
        self.rendering = {content:{reference:self.renderer.emergency_ref},template:false};
    };

    this.build = function(){
        var list = $(self.data.reader.archive()).find("files");
        for(var i=0;i<list.length;i++){
            var content_id = self.execute(function(){return $(list[i]).attr("id");});
            var content = self.data.getArchiveFileItemById(content_id);
            if(content){
                self.addContent(content);
            }
        }
        $(self).trigger("TimelineControllerReady");
    };

    this.unload = function(){
        self.stop();
        self.renderer.clear();
        self.timeline = null;
        self.renderer = null;
        self.data = null;
        self.future_data = null;
    };

    this.update = function(data){
        self.log("<<<<<<< UPDATE >>>>>>");
        if(typeof(data) !== "undefined"){
            self.log("updating xml data .... ");
            self.data = null;
            self.data = data;
        }
        self.timeline = null;
        self.timeline = $(self.data.reader.timeline()).find("content");
        self.build();
        self.prepare_next_day();
        var data_conditions = self.data.getConditions();
        if(data_conditions){
            $(self.conditional).off("ConditionTrigger",self.onConditionTrigger);
            $(self.conditional).on("ConditionTrigger",self.onConditionTrigger);
            self.conditional.update(data_conditions);
        }

				var config = self.data.getConfig();
				var player_mode = config.getMode();
				if(player_mode == 2 || player_mode == 3){
						if(!self.synchronizer){
								self.synchronizer = new SynchronizerController();
								self.conditional.synchronizer = self.synchronizer;
								$(self.synchronizer).on("onSyncChangeMode",self.onSyncModeChanged);
								if(self.synchronizer.using_mode == self.synchronizer.MODE_SLAVE){
										$(self.synchronizer).on("on"+self.name+"SyncRecieve",self.onSyncRecieve);
										$(self.synchronizer).on("on"+self.conditional.name+"SyncRecieve",self.onSyncRecieve);
								}else if(self.synchronizer.using_mode == self.synchronizer.MODE_MASTER){
										$(self.synchronizer).on("onStatusSyncRecieve",self.onStatusSyncRecieve);
								}
						}
						self.synchronizer.update(player_mode,config.getSyncPort(),config.getSyncMachines());
				}
    };

    this.main = function(use_timer){
				var timer_enabled = true;
				if(typeof(use_timer) !== "undefined"){
					timer_enabled = use_timer;
				}
				self.OS.busy = true;
        self.prepared = false;
        //self.log("TIMELINE RENDER ---- Default Content : "+self.renderer.getDefaultDisplayReference()+" To Be Rendered : "+self.scene.content.reference);
        self.rendering = self.renderer.render_scene(self.scene,self.rendering);
				if(self.synchronizer){
						if(self.synchronizer.using_mode == self.synchronizer.MODE_MASTER){
								self.log("Master Send Broadcast Play !! index = "+self.index);
								var broadcast_message = {
									sender:self.name,
									action:{
										do:"play",
										index:self.last_time,
										duration:self.scene.content.content.getDuration(),
										time:self.getDate().getTime(),
										player_id:self.core_id
									}
								};
								self.synchronizer.broadcast(broadcast_message);
								broadcast_message = null;
						}
				}
        self.reportStatus();
        //self.getBenchmark("Render Total Call");

        /// security timer
        clearTimeout(self.security_timer);
				if(timer_enabled){
					var security_time = parseInt(self.scene.content.content.getDuration()/2)*1000;
	        if(security_time < 5000){security_time = 5000;}
					self.log("Activating security timer with delay_time : "+security_time);
	        self.security_timer = setTimeout(self.check_rendering_execution,security_time);
				}
    };

    this.check_rendering_execution = function(){
				self.log("Check for security");
        clearTimeout(self.security_timer);
        if(!self.prepared){
						self.log("Content is NOT prepared");
            self.prepare();
						self.log(self.ERROR['TIMELINE_CONTROLLER']+2,self.name+" : Content "+self.scene.content.reference+" could not load properly... force prepare next content. ");
            self.notifyError(self.ERROR['TIMELINE_CONTROLLER']+2,self.name+" : Content "+self.scene.content.reference+" could not load properly... force prepare next content. ");
        }else{
					self.log("Content IS prepared");
				}
    };

    this.start = function(){
        self.stop();
        self.active = true;
        self.pass = 0;
        self.log("<<<<<<< START >>>>>>");
        self.timeline = $(self.data.reader.timeline()).find("content");
        var date = self.getDate();
        if(self.server_time_activated){
            self.last_time = (date.getUTCHours()*60*60) + (date.getUTCMinutes()*60) + date.getUTCSeconds();
        }else{
            self.last_time = (date.getHours()*60*60) + (date.getMinutes()*60) + date.getSeconds();
        }
        $(self.renderer).on("DisplayControllerElementStarted",self.prepare);
        $(self.renderer).on("DisplayControllerElementError",self.onDisplayControllerElementError);
        if(self.rendering.content.reference == self.renderer.emergency_ref || self.rendering.content.reference == self.renderer.default_ref){
            self.scene = null;
            self.scene = self.renderer.prepare_scene(false,self.renderer.getDefaultDisplayReference(),false);
            self.rendering = self.renderer.render_scene(self.scene,false);
        }else{
            self.prepare();
        }
    };

    this.onSystemTimeUpdate = function(event,before,after){
        self.log("<<<<<<< UPDATE SYSTEM TIME >>>>>>");
        var date = self.getDate();
        if(self.server_time_activated){
            self.last_time = (date.getUTCHours()*60*60) + (date.getUTCMinutes()*60) + date.getUTCSeconds();
        }else{
            self.last_time = (date.getHours()*60*60) + (date.getMinutes()*60) + date.getSeconds();
        }
        self.prepare();
    };

    this.onDisplayControllerElementError = function(e){
        self.scene = null;
        self.scene = self.renderer.prepare_scene(false,self.renderer.getDefaultDisplayReference(),false);
        self.rendering = self.renderer.render_scene(self.scene,self.rendering);
    };

    this.prepare = function(e,sync_time){
				var synchronizer_time = -1;
				if(typeof(sync_time) !== "undefined"){
					synchronizer_time = sync_time;
				}
        clearTimeout(self.security_timer);
        //self.reportStatus();
        if(self.renderer !== null){
            self.toberendered = self.renderer.getDefaultDisplayReference();
        }else{
            self.toberendered = 0;
        }

        self.log("<<<<<<< PREPARE >>>>>>");
        var date = self.getDate();
        var time = 0;
        if(self.server_time_activated){
            time = (date.getUTCHours()*60*60) + (date.getUTCMinutes()*60) + date.getUTCSeconds();
        }else{
            time = (date.getHours()*60*60) + (date.getMinutes()*60) + date.getSeconds();
        }
        time += 1; // round up milliseconds
        var seconds_to_time = new Date(1970,0,1);

        var last = parseInt(self.last_time);
        seconds_to_time.setUTCSeconds(time);
        var now_time = seconds_to_time.getHours()+":"+seconds_to_time.getMinutes()+":"+seconds_to_time.getSeconds();
        self.log("Now : "+now_time+" SoD :"+time);
        seconds_to_time = new Date(1970,0,1);
        seconds_to_time.setUTCSeconds(last);
        var last_time = seconds_to_time.getHours()+":"+seconds_to_time.getMinutes()+":"+seconds_to_time.getSeconds();
        self.log("Last : "+last_time+" SoD :"+last);
        seconds_to_time = new Date(1970,0,1);
        seconds_to_time.setUTCSeconds(self.last_time);
        var in_time = seconds_to_time.getHours()+":"+seconds_to_time.getMinutes()+":"+seconds_to_time.getSeconds();

				if(synchronizer_time >= 0){
					time = synchronizer_time;
				}
        self.line = $(self.timeline).filter(function(i){return $(this).attr('start') > time;}).eq(0);

        var content = self.data.getArchiveFileItemById(self.line.attr("id"));

        var delay = 1;
        if(self.line.length > 0){
            self.last_time = parseInt(self.line.attr("start"));
            delay = self.last_time - time;
        }else{
            self.log("Reached End Of The Day !");
            delay = 86400 - self.last_time;
            self.last_time = 0;
            last = parseInt(self.last_time);
            if(self.future_data && self.timeline.length > 0){
                self.log("Searching next day");
                self.timeline = $(self.future_data.reader.timeline()).find("content");
                self.line = $(self.timeline).filter(function(i){return $(this).attr('start') >= last;}).eq(0);
                if(self.line.length > 0){
                    content = self.future_data.getArchiveFileItemById(self.line.attr("id"));
                }
            }else{
                self.log("Timeline is emtpy or future data could not be loaded.");
                self.notifyError(self.ERROR['TIMELINE_CONTROLLER']+5,self.name+": Timeline is emtpy or future data could not be loaded.");
				delay = self.scene.content.content.getDuration();
				//delay = 0;
            }
        }

				var content_type = "";
        if(content){
            if(content.verify(self.getDate(true).getTime())){
                    if(self.verify_campaign(self.line)){
                        var display_object_reference = self.addContent(content);
                        if(display_object_reference){
                            var DisplayContentElement = self.renderer.getDisplayContentElement(display_object_reference);
														content_type = DisplayContentElement.type;
                            self.renderer.checkAvailability(display_object_reference);
                            if(DisplayContentElement.available && !DisplayContentElement.invalidated){
                                self.toberendered = self.line.attr('id');
                                var timeline_template_id = parseInt(self.line.attr("template_id"));
                                var timeline_duration = parseInt(self.line.attr("duration"));
                                if(!isNaN(timeline_template_id) && timeline_template_id > 0){
                                    var template_element = self.data.getTemplateById(timeline_template_id);
                                    if(template_element){
                                        template_element.loadTemplateContentItems(self.data,self.conditional);
                                        self.template = self.addDisplayTemplate(template_element);
                                    }
                                }else{
                                    self.template = null;
                                    self.template = false;
                                }
                                self.log("Found At -> "+in_time+" SoD:"+self.last_time+" Content : "+self.line.attr("id")+" Duration : "+timeline_duration);
                                self.log("Playing in : "+delay+"s with template : "+self.line.attr("template_id"));
                                self.scene = null;
                                self.scene = self.renderer.prepare_scene(self.template,self.toberendered,false);
                            }else{
																self.scene = null;
                                self.scene = self.renderer.prepare_scene(false,self.toberendered,false);
                                self.log("Content not available yet - Next check in : "+delay+"s");
                            }
                        }else{
														self.scene = null;
                            self.scene = self.renderer.prepare_scene(false,self.toberendered,false);
                            self.log("Unable to create Display Object - Next check in : "+delay+"s");
                        }
                    }else{
												self.scene = null;
                        self.scene = self.renderer.prepare_scene(false,self.toberendered,false);
                        self.log("Campaign Expired - Next check in : "+delay+"s");
                        self.notifyError(self.ERROR['TIMELINE_CONTROLLER']+3,self.name+": Campaign Expired: Content "+self.line.attr("id")+" is out of campaign range.");
                    }
            }else{
								self.scene = null;
                self.scene = self.renderer.prepare_scene(false,self.toberendered,false);
                self.log("Content Expired or Not Active yet - Next check in : "+delay+"s");
                self.notifyError(self.ERROR['TIMELINE_CONTROLLER']+4,self.name+": Content Expired or Not Active yet: Content "+self.line.attr("id")+" is out of date range.");
            }
        }else{
						self.scene = null;
            self.scene = self.renderer.prepare_scene(false,self.toberendered,false);
            self.log("No content found - Next check in : "+delay+"s");
        }
		clearTimeout(self.timer);
		if(synchronizer_time == -1){
			if(self.platform == "SOC" && delay < 4){
				self.log("Delay is lower then 4s : re-prepare");
				self.timer = setTimeout(self.main,4000);
			}else if(delay < 2){
				self.log("Delay is lower then 2s : re-prepare");
				self.timer = setTimeout(self.main,2000);
			}else{
				self.log("Delay to next render : "+delay);
				self.timer = setTimeout(self.main,delay*1000);
			}
		}
		self.OS.busy = false;
		self.prepared = true;

		if(self.synchronizer){
				if(self.synchronizer.using_mode == self.synchronizer.MODE_MASTER){
						self.log("Master Send Broadcast Preload !! Actual index :"+self.index);
						var broadcast_message = {
							sender:self.name,
							action:{
								do:"preload",
								index:time,
								time:self.getDate().getTime(),
								player_id:self.core_id
							}
						};
						self.synchronizer.broadcast(broadcast_message);
						broadcast_message = null;
				}
		}

        var x = $(self.timeline).filter(function(i){return $(this).attr('position') == self.line.attr("position")-1;}).eq(0);
        if(delay > (parseInt(x.attr("duration"))+4)){
            var rendering2 = {content:{reference:self.renderer.default_ref},template:false};
            var scene2 = self.renderer.prepare_scene(false,self.renderer.default_ref,false);
            rendering2 = self.renderer.render_scene(scene2);
            $(self.renderer).off("DisplayControllerElementStarted");
        }
    };

    this.stop = function(){
        self.active = false;
        self.log("<<<<<<< STOP >>>>>>");
        clearTimeout(self.timer);
				clearTimeout(self.security_timer);
        $(self.renderer).off("DisplayControllerElementStarted");
        $(self.renderer).off("DisplayControllerElementError");
    };

    this.addContent = function(ContentElement){
        var displayObjectRef = false;
        displayObjectRef = self.renderer.addDisplayElement(ContentElement);
        if(!displayObjectRef){
            self.log("no display object for content : "+ContentElement.getSource());
            self.notifyError(self.ERROR['TIMELINE_CONTROLLER']+1,self.name+": Function addContent() failed : no display object for content : "+ContentElement.getSource());
        }else{
					self.renderer.checkAvailability(displayObjectRef);
				}
        return displayObjectRef;
    };

    this.addDisplayTemplate = function(TemplateElement){
        var id = 0;
        if(TemplateElement){
            var id = self.execute(function(){return TemplateElement.getId();});
            if(id){
                id = parseInt(id);
            }
        }
        if(id > 0){
            if(typeof(self.renderer.template_elements[id]) !== "undefined"){
                // TODO : IMPROVE
                var prev = ""+$(self.renderer.template_elements[id].data.data)[0].outerHTML;
                var post = ""+$(TemplateElement.data)[0].outerHTML;
                if(prev != post){
                    //self.renderer.template.unload(id); // OLD TODO: review template.unload()
                    self.renderer.template.unload();
                    self.renderer.template_elements[id].data = TemplateElement;
                    self.renderer.template.load();
                }
            }else{
                self.renderer.template_elements[id] = null;
                self.renderer.template_elements[id] = new DisplayTemplate(TemplateElement,self.renderer,self.conditional);
            }
        }
        return id;
    };

    this.jump = function(index){
        self.log("No action jump available");
    };

    this.reportStatus = function(){
        var campaign_id = $(self.data.reader.timeline()).find("content[start='"+self.last_time+"']").attr("campaign_id");
        var stats_conditions = "";
        if(self.conditional){
            stats_conditions = self.conditional.createConditionsStat(self.data.getStats().getConditionStats());
        }
        self.OS.id_content_now = self.rendering.content.reference;
        self.data.reportStatistics(self.rendering.content.reference,campaign_id,stats_conditions);
        //self.OS.writeFile("admira.status",self.rendering.content.reference+":"+self.pass+":"+self.playlist_id+":"+self.version);
    };

    this.verify_campaign = function(line){
        var is_active = true;
        var campaign_id = $(line).attr('campaign_id');
        if(campaign_id){
            if(campaign_id != "" && campaign_id != "0"){
                var expiration = self.getDate(self.server_time_activated,2147483647000);
                var temp_expiration = $(line).attr('expiration');
                var day = parseInt(temp_expiration.split("-")[0]);
                var month = parseInt(temp_expiration.split("-")[1])-1;
                var year = parseInt(temp_expiration.split("-")[2]);
                if(self.server_time_activated){
                    expiration.setUTCFullYear(year);
                    expiration.setUTCMonth(month);
                    expiration.setUTCDate(day);
                }else{
                    expiration.setFullYear(year);
                    expiration.setMonth(month);
                    expiration.setDate(day);
                }
                expiration = expiration.getTime()+86399000; // Add the 86400 seconds of the entire day -1sec
                var now = self.getDate().getTime();
                if(isNaN(expiration)){
                    is_active = true;
                }else{
                    if(expiration > now){
                        is_active = true;
                    }else{
                        is_active = false;
                    }
                }
            }
        }
        return is_active;
    };

    this.prepare_next_day = function(){
        var dow = self.getDate().getDay();
        dow+=1;
        if(dow == 0) dow = 7;
        if(dow == 8) dow = 1;

        var url = self.rootDir+self.ROUTES['XML'].path+dow+".xml";
        if(self.platform == "CHROMEOS"){
              url = self.OS.fs.root.toURL()+self.rootDir+self.ROUTES['XML'].path+dow+".xml";
        }else{
              url = self.OS.rootDir+self.ROUTES['XML'].path+dow+".xml";
        }
        self.future_data = null;
        self.future_data = new DataManagerController();
        self.future_data.load(url,-1);
        $(self.future_data).one("DataManagerControllerDataLoaded",self.onDataLoaded);
    };

    this.onDataLoaded = function(){
        self.log("Next day ready");
    };

    this.onConditionTrigger = function(e,interruption,action,action_value){
        self.log("Prepare Trigger Action : "+action+" Value : "+action_value+" Interrupts : "+interruption);
    };

		this.onSyncRecieve = function(event,action){
				if(action.player_id == ""+self.synchronizer.master_server_id){
					if(action.do == "play"){
							self.log("<#> Play ! Index = "+self.index);
							self.main(false);
					}else if(action.do == "preload"){
							self.log("<#> Actual Index = "+self.index+" Want to preload index = "+action.index);
							if(action.index != self.index){
									self.log("<#> Index updated to : "+action.index);
									self.index = parseInt(action.index);
							}
							self.log("<#> Checknext for: "+action.index);
							self.prepare(false,action.index);
					}else if(action.do == "condition"){
							self.log("<#> Set condition from MASTER");
							self.conditional.onNewConditionValue(action.condition,false);
					}
				}
				if(!self.active && self.synchronizer.using_mode == self.synchronizer.MODE_MASTER){
						self.start();
				}else if(self.active){
						self.log("I'm SLAVE stop autoplaylist playback");
						self.stop();
				}
		};

		this.onStatusSyncRecieve = function(event,action){
				if(!self.active && self.synchronizer.using_mode == self.synchronizer.MODE_MASTER){
						self.start();
				}
		};

		this.onSyncModeChanged = function(){
				self.log("Sync Mode Changed to : "+self.synchronizer.using_mode);
				$(self.synchronizer).off("on"+self.name+"SyncRecieve",self.onSyncRecieve);
				$(self.synchronizer).off("on"+self.conditional.name+"SyncRecieve",self.onSyncRecieve);
				$(self.synchronizer).off("onStatusSyncRecieve",self.onStatusSyncRecieve);

				if(self.synchronizer.using_mode == self.synchronizer.MODE_SLAVE){
						self.stop();
						$(self.synchronizer).on("on"+self.name+"SyncRecieve",self.onSyncRecieve);
						$(self.synchronizer).on("on"+self.conditional.name+"SyncRecieve",self.onSyncRecieve);
				}else if(self.synchronizer.using_mode == self.synchronizer.MODE_MASTER){
						$(self.synchronizer).on("onStatusSyncRecieve",self.onStatusSyncRecieve);
						self.stop();
						self.prepare();
						self.start();
				}
		};
}
TimelineController.prototype = new SystemController();
